// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/file_analysis/analyzer/x509/types.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_x509_types_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_x509_types_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_x509_types_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifType { namespace Record{ namespace X509 {  extern zeek::IntrusivePtr<zeek::RecordType> Certificate;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace X509 {  extern zeek::IntrusivePtr<zeek::RecordType> Extension;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace X509 {  extern zeek::IntrusivePtr<zeek::RecordType> BasicConstraints;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace X509 {  extern zeek::IntrusivePtr<zeek::RecordType> SubjectAlternativeName;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace X509 {  extern zeek::IntrusivePtr<zeek::RecordType> Result;  } } }}

#endif
