// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/input/input.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_input_input_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_input_input_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_input_input_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace BifEnum { namespace Input { enum Event {
	EVENT_NEW = 0,
	EVENT_CHANGED = 1,
	EVENT_REMOVED = 2,
}; } }
namespace zeek { namespace BifType { namespace Enum{ namespace Input {  extern zeek::IntrusivePtr<zeek::EnumType> Event;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Input {  extern zeek::IntrusivePtr<zeek::RecordType> TableDescription;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Input {  extern zeek::IntrusivePtr<zeek::RecordType> EventDescription;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Input {  extern zeek::IntrusivePtr<zeek::RecordType> AnalysisDescription;  } } }}
namespace zeek { namespace BifFunc { namespace Input { extern zeek::detail::BifReturnVal __create_table_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Input { extern zeek::detail::BifReturnVal __create_event_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Input { extern zeek::detail::BifReturnVal __create_analysis_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Input { extern zeek::detail::BifReturnVal __remove_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Input { extern zeek::detail::BifReturnVal __force_update_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifConst { namespace Input {  extern int accept_unsupported_types;  }  } }

#endif
