// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/analyzer.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_analyzer_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_analyzer_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_analyzer_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __enable_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __disable_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __disable_all_analyzers_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __register_for_port_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __schedule_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __name_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __tag_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Analyzer { extern zeek::detail::BifReturnVal __has_tag_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
