// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/broker/messaging.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_messaging_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_messaging_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_messaging_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifType { namespace Record{ namespace Broker {  extern zeek::IntrusivePtr<zeek::RecordType> Event;  } } }}
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal make_event_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal publish_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __flush_logs_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __publish_id_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __auto_publish_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __auto_unpublish_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __subscribe_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __forward_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __unsubscribe_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifType { namespace Record{ namespace Cluster {  extern zeek::IntrusivePtr<zeek::RecordType> Pool;  } } }}
namespace zeek { namespace BifFunc { namespace Cluster { extern zeek::detail::BifReturnVal publish_rr_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Cluster { extern zeek::detail::BifReturnVal publish_hrw_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
