// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/dce-rpc/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_dce_rpc_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_dce_rpc_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_dce_rpc_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr dce_rpc_message; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_message(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek_uint_t fid, zeek_uint_t ptype_id, zeek::IntrusivePtr<zeek::Val> ptype);  }  }
extern zeek::EventHandlerPtr dce_rpc_bind; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_bind(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek_uint_t ctx_id, zeek::IntrusivePtr<zeek::StringVal> uuid, zeek_uint_t ver_major, zeek_uint_t ver_minor);  }  }
extern zeek::EventHandlerPtr dce_rpc_alter_context; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_alter_context(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek_uint_t ctx_id, zeek::IntrusivePtr<zeek::StringVal> uuid, zeek_uint_t ver_major, zeek_uint_t ver_minor);  }  }
extern zeek::EventHandlerPtr dce_rpc_bind_ack; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_bind_ack(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek::IntrusivePtr<zeek::StringVal> sec_addr);  }  }
extern zeek::EventHandlerPtr dce_rpc_alter_context_resp; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_alter_context_resp(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid);  }  }
extern zeek::EventHandlerPtr dce_rpc_request; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek_uint_t ctx_id, zeek_uint_t opnum, zeek_uint_t stub_len);  }  }
extern zeek::EventHandlerPtr dce_rpc_response; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_response(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek_uint_t ctx_id, zeek_uint_t opnum, zeek_uint_t stub_len);  }  }
extern zeek::EventHandlerPtr dce_rpc_request_stub; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_request_stub(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek_uint_t ctx_id, zeek_uint_t opnum, zeek::IntrusivePtr<zeek::StringVal> stub);  }  }
extern zeek::EventHandlerPtr dce_rpc_response_stub; 
namespace zeek { namespace BifEvent {  void enqueue_dce_rpc_response_stub(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t fid, zeek_uint_t ctx_id, zeek_uint_t opnum, zeek::IntrusivePtr<zeek::StringVal> stub);  }  }

#endif
