// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/http/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_http_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_http_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_http_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr http_request; 
namespace zeek { namespace BifEvent {  void enqueue_http_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> method, zeek::IntrusivePtr<zeek::StringVal> original_URI, zeek::IntrusivePtr<zeek::StringVal> unescaped_URI, zeek::IntrusivePtr<zeek::StringVal> version);  }  }
extern zeek::EventHandlerPtr http_reply; 
namespace zeek { namespace BifEvent {  void enqueue_http_reply(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> version, zeek_uint_t code, zeek::IntrusivePtr<zeek::StringVal> reason);  }  }
extern zeek::EventHandlerPtr http_header; 
namespace zeek { namespace BifEvent {  void enqueue_http_header(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> original_name, zeek::IntrusivePtr<zeek::StringVal> name, zeek::IntrusivePtr<zeek::StringVal> value);  }  }
extern zeek::EventHandlerPtr http_all_headers; 
namespace zeek { namespace BifEvent {  void enqueue_http_all_headers(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::Val> hlist);  }  }
extern zeek::EventHandlerPtr http_begin_entity; 
namespace zeek { namespace BifEvent {  void enqueue_http_begin_entity(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig);  }  }
extern zeek::EventHandlerPtr http_end_entity; 
namespace zeek { namespace BifEvent {  void enqueue_http_end_entity(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig);  }  }
extern zeek::EventHandlerPtr http_entity_data; 
namespace zeek { namespace BifEvent {  void enqueue_http_entity_data(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek_uint_t length, zeek::IntrusivePtr<zeek::StringVal> data);  }  }
extern zeek::EventHandlerPtr http_content_type; 
namespace zeek { namespace BifEvent {  void enqueue_http_content_type(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> ty, zeek::IntrusivePtr<zeek::StringVal> subty);  }  }
extern zeek::EventHandlerPtr http_message_done; 
namespace zeek { namespace BifEvent {  void enqueue_http_message_done(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::Val> stat);  }  }
extern zeek::EventHandlerPtr http_event; 
namespace zeek { namespace BifEvent {  void enqueue_http_event(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> event_type, zeek::IntrusivePtr<zeek::StringVal> detail);  }  }
extern zeek::EventHandlerPtr http_stats; 
namespace zeek { namespace BifEvent {  void enqueue_http_stats(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> stats);  }  }
extern zeek::EventHandlerPtr http_connection_upgrade; 
namespace zeek { namespace BifEvent {  void enqueue_http_connection_upgrade(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> protocol);  }  }

#endif
