// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/finger/legacy/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_finger_legacy_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_finger_legacy_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_finger_legacy_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr finger_request; 
namespace zeek { namespace BifEvent {  void enqueue_finger_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int full, zeek::IntrusivePtr<zeek::StringVal> username, zeek::IntrusivePtr<zeek::StringVal> hostname);  }  }
extern zeek::EventHandlerPtr finger_reply; 
namespace zeek { namespace BifEvent {  void enqueue_finger_reply(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> reply_line);  }  }

#endif
