// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/file_analysis/analyzer/pe/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_pe_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_pe_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_pe_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr pe_dos_header; 
namespace zeek { namespace BifEvent {  void enqueue_pe_dos_header(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::Val> h);  }  }
extern zeek::EventHandlerPtr pe_dos_code; 
namespace zeek { namespace BifEvent {  void enqueue_pe_dos_code(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::StringVal> code);  }  }
extern zeek::EventHandlerPtr pe_file_header; 
namespace zeek { namespace BifEvent {  void enqueue_pe_file_header(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::Val> h);  }  }
extern zeek::EventHandlerPtr pe_optional_header; 
namespace zeek { namespace BifEvent {  void enqueue_pe_optional_header(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::Val> h);  }  }
extern zeek::EventHandlerPtr pe_section_header; 
namespace zeek { namespace BifEvent {  void enqueue_pe_section_header(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::Val> h);  }  }

#endif
