// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/smb/smb1_com_read_andx.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_smb_smb1_com_read_andx_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_smb_smb1_com_read_andx_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_smb_smb1_com_read_andx_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr smb1_read_andx_request; 
namespace zeek { namespace BifEvent {  void enqueue_smb1_read_andx_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek_uint_t file_id, zeek_uint_t offset, zeek_uint_t length);  }  }
extern zeek::EventHandlerPtr smb1_read_andx_response; 
namespace zeek { namespace BifEvent {  void enqueue_smb1_read_andx_response(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek_uint_t data_len);  }  }

#endif
