// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/smb/smb2_com_set_info.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_smb_smb2_com_set_info_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_smb_smb2_com_set_info_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_smb_smb2_com_set_info_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr smb2_file_rename; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_rename(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek::IntrusivePtr<zeek::StringVal> dst_filename);  }  }
extern zeek::EventHandlerPtr smb2_file_delete; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_delete(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, int delete_pending);  }  }
extern zeek::EventHandlerPtr smb2_file_sattr; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_sattr(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek::IntrusivePtr<zeek::Val> times, zeek::IntrusivePtr<zeek::Val> attrs);  }  }
extern zeek::EventHandlerPtr smb2_file_allocation; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_allocation(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_int_t alloc_size);  }  }
extern zeek::EventHandlerPtr smb2_file_endoffile; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_endoffile(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_int_t end_of_file);  }  }
extern zeek::EventHandlerPtr smb2_file_mode; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_mode(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_uint_t mode);  }  }
extern zeek::EventHandlerPtr smb2_file_pipe; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_pipe(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_uint_t read_mode, zeek_uint_t completion_mode);  }  }
extern zeek::EventHandlerPtr smb2_file_position; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_position(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_int_t current_byte_offset);  }  }
extern zeek::EventHandlerPtr smb2_file_shortname; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_shortname(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek::IntrusivePtr<zeek::StringVal> file_name);  }  }
extern zeek::EventHandlerPtr smb2_file_validdatalength; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_validdatalength(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_int_t valid_data_length);  }  }
extern zeek::EventHandlerPtr smb2_file_fullea; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_fullea(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek::IntrusivePtr<zeek::Val> file_eas);  }  }
extern zeek::EventHandlerPtr smb2_file_link; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_link(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek_uint_t root_directory, zeek::IntrusivePtr<zeek::StringVal> file_name);  }  }
extern zeek::EventHandlerPtr smb2_file_fscontrol; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_fscontrol(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek::IntrusivePtr<zeek::Val> fs_control);  }  }
extern zeek::EventHandlerPtr smb2_file_fsobjectid; 
namespace zeek { namespace BifEvent {  void enqueue_smb2_file_fsobjectid(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> file_id, zeek::IntrusivePtr<zeek::Val> object_id, zeek::IntrusivePtr<zeek::StringVal> extended_info);  }  }

#endif
