// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/broker/store.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_store_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_store_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_store_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifType { namespace Record{ namespace Broker {  extern zeek::IntrusivePtr<zeek::RecordType> QueryResult;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Broker {  extern zeek::IntrusivePtr<zeek::RecordType> BackendOptions;  } } }}
namespace BifEnum { namespace Broker { enum BackendType {
	MEMORY,
	SQLITE,
}; } }
namespace zeek { namespace BifType { namespace Enum{ namespace Broker {  extern zeek::IntrusivePtr<zeek::EnumType> BackendType;  } } }}
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __create_master_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __create_clone_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __is_closed_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __close_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __store_name_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __exists_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __get_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __put_unique_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __get_index_from_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __keys_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __put_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __erase_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __increment_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __decrement_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __append_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __insert_into_set_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __insert_into_table_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __remove_from_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __push_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __pop_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
