// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/conn-size/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_conn_size_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_conn_size_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_conn_size_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr conn_bytes_threshold_crossed; 
namespace zeek { namespace BifEvent {  void enqueue_conn_bytes_threshold_crossed(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t threshold, int is_orig);  }  }
extern zeek::EventHandlerPtr conn_packets_threshold_crossed; 
namespace zeek { namespace BifEvent {  void enqueue_conn_packets_threshold_crossed(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t threshold, int is_orig);  }  }
extern zeek::EventHandlerPtr conn_duration_threshold_crossed; 
namespace zeek { namespace BifEvent {  void enqueue_conn_duration_threshold_crossed(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, double threshold, int is_orig);  }  }

#endif
