// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/file_analysis/analyzer/hash/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_hash_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_hash_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_analyzer_hash_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr file_hash; 
namespace zeek { namespace BifEvent {  void enqueue_file_hash(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::StringVal> kind, zeek::IntrusivePtr<zeek::StringVal> hash);  }  }

#endif
