// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/probabilistic/cardinality-counter.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_cardinality_counter_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_cardinality_counter_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_cardinality_counter_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal hll_cardinality_init_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal hll_cardinality_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal hll_cardinality_merge_into_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal hll_cardinality_estimate_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal hll_cardinality_copy_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }

#endif
