// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/packet_analysis/protocol/udp/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_udp_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_udp_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_udp_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr udp_request; 
namespace zeek { namespace BifEvent {  void enqueue_udp_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* u);  }  }
extern zeek::EventHandlerPtr udp_reply; 
namespace zeek { namespace BifEvent {  void enqueue_udp_reply(zeek::analyzer::Analyzer* analyzer, zeek::Connection* u);  }  }
extern zeek::EventHandlerPtr udp_contents; 
namespace zeek { namespace BifEvent {  void enqueue_udp_contents(zeek::analyzer::Analyzer* analyzer, zeek::Connection* u, int is_orig, zeek::IntrusivePtr<zeek::StringVal> contents);  }  }
extern zeek::EventHandlerPtr udp_multiple_checksum_errors; 
namespace zeek { namespace BifEvent {  void enqueue_udp_multiple_checksum_errors(zeek::analyzer::Analyzer* analyzer, zeek::Connection* u, int is_orig, zeek_uint_t threshold);  }  }

#endif
