// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/logging/logging.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_logging_logging_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_logging_logging_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_logging_logging_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifType { namespace Record{ namespace Log {  extern zeek::IntrusivePtr<zeek::RecordType> Filter;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Log {  extern zeek::IntrusivePtr<zeek::RecordType> Stream;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Log {  extern zeek::IntrusivePtr<zeek::RecordType> RotationInfo;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Log {  extern zeek::IntrusivePtr<zeek::RecordType> RotationFmtInfo;  } } }}
namespace BifEnum { namespace Log { enum PrintLogType {
	REDIRECT_NONE,
	REDIRECT_STDOUT,
	REDIRECT_ALL,
}; } }
namespace zeek { namespace BifType { namespace Enum{ namespace Log {  extern zeek::IntrusivePtr<zeek::EnumType> PrintLogType;  } } }}
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __create_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __remove_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __enable_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __disable_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __add_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __remove_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __write_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __set_buf_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Log { extern zeek::detail::BifReturnVal __flush_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
