// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/file_analysis/file_analysis.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_file_analysis_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_file_analysis_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_file_analysis_file_analysis_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifType { namespace Record{ namespace Files {  extern zeek::IntrusivePtr<zeek::RecordType> AnalyzerArgs;  } } }}
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __set_timeout_interval_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __enable_reassembly_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __disable_reassembly_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __set_reassembly_buffer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __enable_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __disable_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __analyzer_enabled_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __add_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __remove_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __stop_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __analyzer_name_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __file_exists_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Files { extern zeek::detail::BifReturnVal __lookup_file_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal set_file_handle_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }

#endif
