// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/radius/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_radius_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_radius_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_radius_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr radius_message; 
namespace zeek { namespace BifEvent {  void enqueue_radius_message(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> result);  }  }
extern zeek::EventHandlerPtr radius_attribute; 
namespace zeek { namespace BifEvent {  void enqueue_radius_attribute(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t attr_type, zeek::IntrusivePtr<zeek::StringVal> value);  }  }
namespace zeek { namespace BifType { namespace Vector{ namespace RADIUS {  extern zeek::IntrusivePtr<zeek::VectorType> AttributeList;  } } }}
namespace zeek { namespace BifType { namespace Table{ namespace RADIUS {  extern zeek::IntrusivePtr<zeek::TableType> Attributes;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace RADIUS {  extern zeek::IntrusivePtr<zeek::RecordType> Message;  } } }}

#endif
