# This file was automatically generated by SWIG (https://www.swig.org).
# Version 4.2.1
#
# Do not make changes to this file unless you know what you are doing - modify
# the SWIG interface file instead.

from sys import version_info as _swig_python_version_info
# Import the low-level C/C++ module
if __package__ or "." in __name__:
    from . import _SubnetTree
else:
    import _SubnetTree

try:
    import builtins as __builtin__
except ImportError:
    import __builtin__

def _swig_repr(self):
    try:
        strthis = "proxy of " + self.this.__repr__()
    except __builtin__.Exception:
        strthis = ""
    return "<%s.%s; %s >" % (self.__class__.__module__, self.__class__.__name__, strthis,)


def _swig_setattr_nondynamic_instance_variable(set):
    def set_instance_attr(self, name, value):
        if name == "this":
            set(self, name, value)
        elif name == "thisown":
            self.this.own(value)
        elif hasattr(self, name) and isinstance(getattr(type(self), name), property):
            set(self, name, value)
        else:
            raise AttributeError("You cannot add instance attributes to %s" % self)
    return set_instance_attr


def _swig_setattr_nondynamic_class_variable(set):
    def set_class_attr(cls, name, value):
        if hasattr(cls, name) and not isinstance(getattr(cls, name), property):
            set(cls, name, value)
        else:
            raise AttributeError("You cannot add class attributes to %s" % cls)
    return set_class_attr


def _swig_add_metaclass(metaclass):
    """Class decorator for adding a metaclass to a SWIG wrapped class - a slimmed down version of six.add_metaclass"""
    def wrapper(cls):
        return metaclass(cls.__name__, cls.__bases__, cls.__dict__.copy())
    return wrapper


class _SwigNonDynamicMeta(type):
    """Meta class to enforce nondynamic attributes (no new attributes) for a class"""
    __setattr__ = _swig_setattr_nondynamic_class_variable(type.__setattr__)


class inx_addr(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr
    sin = property(_SubnetTree.inx_addr_sin_get, _SubnetTree.inx_addr_sin_set)
    sin6 = property(_SubnetTree.inx_addr_sin6_get, _SubnetTree.inx_addr_sin6_set)

    def __init__(self):
        _SubnetTree.inx_addr_swiginit(self, _SubnetTree.new_inx_addr())
    __swig_destroy__ = _SubnetTree.delete_inx_addr

# Register inx_addr in _SubnetTree:
_SubnetTree.inx_addr_swigregister(inx_addr)
class SubnetTree(object):
    thisown = property(lambda x: x.this.own(), lambda x, v: x.this.own(v), doc="The membership flag")
    __repr__ = _swig_repr

    def __init__(self, binary_lookup_mode=False):
        _SubnetTree.SubnetTree_swiginit(self, _SubnetTree.new_SubnetTree(binary_lookup_mode))
    __swig_destroy__ = _SubnetTree.delete_SubnetTree

    def insert(self, *args):
        return _SubnetTree.SubnetTree_insert(self, *args)

    def remove(self, *args):
        return _SubnetTree.SubnetTree_remove(self, *args)

    def lookup(self, *args):
        return _SubnetTree.SubnetTree_lookup(self, *args)

    def prefixes(self, ipv4_native=False, with_len=True):
        return _SubnetTree.SubnetTree_prefixes(self, ipv4_native, with_len)

    def get_binary_lookup_mode(self):
        return _SubnetTree.SubnetTree_get_binary_lookup_mode(self)

    def set_binary_lookup_mode(self, binary_lookup_mode=True):
        return _SubnetTree.SubnetTree_set_binary_lookup_mode(self, binary_lookup_mode)

    def __contains__(self, *args):
        return _SubnetTree.SubnetTree___contains__(self, *args)

    def __getitem__(self, cidr):
        return _SubnetTree.SubnetTree___getitem__(self, cidr)

    def __setitem__(self, cidr, data):
        return _SubnetTree.SubnetTree___setitem__(self, cidr, data)

    def __delitem__(self, cidr):
        return _SubnetTree.SubnetTree___delitem__(self, cidr)

# Register SubnetTree in _SubnetTree:
_SubnetTree.SubnetTree_swigregister(SubnetTree)

