// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/iosource/pcap/pcap.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_iosource_pcap_pcap_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_iosource_pcap_pcap_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_iosource_pcap_pcap_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifConst { namespace Pcap {  extern zeek_uint_t snaplen;  }  } }
namespace zeek { namespace BifConst { namespace Pcap {  extern zeek_uint_t bufsize;  }  } }
namespace zeek { namespace BifConst { namespace Pcap {  extern double non_fd_timeout;  }  } }
namespace zeek { namespace BifFunc { namespace Pcap { extern zeek::detail::BifReturnVal precompile_pcap_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Pcap { extern zeek::detail::BifReturnVal install_pcap_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Pcap { extern zeek::detail::BifReturnVal error_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Pcap { extern zeek::detail::BifReturnVal get_filter_state_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Pcap { extern zeek::detail::BifReturnVal get_filter_state_string_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Pcap { extern zeek::detail::BifReturnVal findalldevs_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
