#pragma once

// Headers to include by generated BiF code.
#include "zeek/BifReturnVal.h"
#include "zeek/Conn.h"
#include "zeek/Event.h"
#include "zeek/EventRegistry.h"
#include "zeek/ID.h"
#include "zeek/NetVar.h"
#include "zeek/Reporter.h"
