// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/input/readers/ascii/ascii.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_input_readers_ascii_ascii_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_input_readers_ascii_ascii_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_input_readers_ascii_ascii_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifConst { namespace InputAscii {  extern zeek::IntrusivePtr<zeek::StringVal> separator;  }  } }
namespace zeek { namespace BifConst { namespace InputAscii {  extern zeek::IntrusivePtr<zeek::StringVal> set_separator;  }  } }
namespace zeek { namespace BifConst { namespace InputAscii {  extern zeek::IntrusivePtr<zeek::StringVal> empty_field;  }  } }
namespace zeek { namespace BifConst { namespace InputAscii {  extern zeek::IntrusivePtr<zeek::StringVal> unset_field;  }  } }
namespace zeek { namespace BifConst { namespace InputAscii {  extern int fail_on_invalid_lines;  }  } }
namespace zeek { namespace BifConst { namespace InputAscii {  extern int fail_on_file_problem;  }  } }
namespace zeek { namespace BifConst { namespace InputAscii {  extern zeek::IntrusivePtr<zeek::StringVal> path_prefix;  }  } }

#endif
