// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/telemetry/telemetry.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_telemetry_telemetry_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_telemetry_telemetry_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_telemetry_telemetry_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace BifEnum { namespace Telemetry { enum MetricType {
	DOUBLE_COUNTER,
	INT_COUNTER,
	DOUBLE_GAUGE,
	INT_GAUGE,
	DOUBLE_HISTOGRAM,
	INT_HISTOGRAM,
}; } }
namespace zeek { namespace BifType { namespace Enum{ namespace Telemetry {  extern zeek::IntrusivePtr<zeek::EnumType> MetricType;  } } }}
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_counter_family_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_counter_metric_get_or_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_counter_inc_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_counter_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_counter_family_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_counter_metric_get_or_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_counter_inc_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_counter_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_gauge_family_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_gauge_metric_get_or_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_gauge_inc_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_gauge_dec_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_gauge_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_gauge_family_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_gauge_metric_get_or_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_gauge_inc_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_gauge_dec_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_gauge_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_histogram_family_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_histogram_metric_get_or_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_histogram_observe_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __int_histogram_sum_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_histogram_family_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_histogram_metric_get_or_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_histogram_observe_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __dbl_histogram_sum_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __collect_metrics_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Telemetry { extern zeek::detail::BifReturnVal __collect_histogram_metrics_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
