// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/analyzer/protocol/gnutella/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_gnutella_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_gnutella_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_analyzer_protocol_gnutella_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr gnutella_text_msg; 
namespace zeek { namespace BifEvent {  void enqueue_gnutella_text_msg(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int orig, zeek::IntrusivePtr<zeek::StringVal> headers);  }  }
extern zeek::EventHandlerPtr gnutella_binary_msg; 
namespace zeek { namespace BifEvent {  void enqueue_gnutella_binary_msg(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int orig, zeek_uint_t msg_type, zeek_uint_t ttl, zeek_uint_t hops, zeek_uint_t msg_len, zeek::IntrusivePtr<zeek::StringVal> payload, zeek_uint_t payload_len, int trunc, int complete);  }  }
extern zeek::EventHandlerPtr gnutella_partial_binary_msg; 
namespace zeek { namespace BifEvent {  void enqueue_gnutella_partial_binary_msg(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int orig, zeek::IntrusivePtr<zeek::StringVal> msg, zeek_uint_t len);  }  }
extern zeek::EventHandlerPtr gnutella_establish; 
namespace zeek { namespace BifEvent {  void enqueue_gnutella_establish(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }  }
extern zeek::EventHandlerPtr gnutella_not_establish; 
namespace zeek { namespace BifEvent {  void enqueue_gnutella_not_establish(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }  }
extern zeek::EventHandlerPtr gnutella_http_notify; 
namespace zeek { namespace BifEvent {  void enqueue_gnutella_http_notify(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }  }

#endif
