// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/probabilistic/bloom-filter.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_bloom_filter_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_bloom_filter_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_bloom_filter_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_basic_init_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_basic_init2_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_counting_init_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_decrement_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_merge_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_intersect_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal bloomfilter_internal_state_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }

#endif
