// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/packet_analysis/protocol/teredo/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_teredo_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_teredo_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_teredo_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr teredo_packet; 
namespace zeek { namespace BifEvent {  void enqueue_teredo_packet(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }  }
extern zeek::EventHandlerPtr teredo_authentication; 
namespace zeek { namespace BifEvent {  void enqueue_teredo_authentication(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }  }
extern zeek::EventHandlerPtr teredo_origin_indication; 
namespace zeek { namespace BifEvent {  void enqueue_teredo_origin_indication(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }  }
extern zeek::EventHandlerPtr teredo_bubble; 
namespace zeek { namespace BifEvent {  void enqueue_teredo_bubble(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }  }

#endif
