// See the file "COPYING" in the main distribution directory for copyright.

#pragma once

#define ZEEK_SCRIPT_INSTALL_PATH "/usr/local/share/zeek"
#define BRO_PLUGIN_INSTALL_PATH "/usr/local/lib/zeek/plugins"
#define ZEEK_PLUGIN_INSTALL_PATH "/usr/local/lib/zeek/plugins"
#define DEFAULT_ZEEKPATH ".:/usr/local/share/zeek:/usr/local/share/zeek/policy:/usr/local/share/zeek/site:/usr/local/share/zeek/builtin-plugins"
#define ZEEK_SPICY_MODULE_PATH ""
#define ZEEK_SPICY_LIBRARY_PATH ""
#define ZEEK_SPICY_DATA_PATH ""
