// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/packet_analysis/protocol/teredo/functions.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_teredo_functions_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_teredo_functions_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_teredo_functions_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifFunc { namespace PacketAnalyzer::TEREDO { extern zeek::detail::BifReturnVal remove_teredo_connection_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
