// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/packet_analysis/protocol/arp/events.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_arp_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_arp_events_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_packet_analysis_protocol_arp_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr arp_request; 
namespace zeek { namespace BifEvent {  void enqueue_arp_request(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::StringVal> mac_src, zeek::IntrusivePtr<zeek::StringVal> mac_dst, zeek::IntrusivePtr<zeek::AddrVal> SPA, zeek::IntrusivePtr<zeek::StringVal> SHA, zeek::IntrusivePtr<zeek::AddrVal> TPA, zeek::IntrusivePtr<zeek::StringVal> THA);  }  }
extern zeek::EventHandlerPtr arp_reply; 
namespace zeek { namespace BifEvent {  void enqueue_arp_reply(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::StringVal> mac_src, zeek::IntrusivePtr<zeek::StringVal> mac_dst, zeek::IntrusivePtr<zeek::AddrVal> SPA, zeek::IntrusivePtr<zeek::StringVal> SHA, zeek::IntrusivePtr<zeek::AddrVal> TPA, zeek::IntrusivePtr<zeek::StringVal> THA);  }  }
extern zeek::EventHandlerPtr bad_arp; 
namespace zeek { namespace BifEvent {  void enqueue_bad_arp(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::AddrVal> SPA, zeek::IntrusivePtr<zeek::StringVal> SHA, zeek::IntrusivePtr<zeek::AddrVal> TPA, zeek::IntrusivePtr<zeek::StringVal> THA, zeek::IntrusivePtr<zeek::StringVal> explanation);  }  }

#endif
