// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/broker/data.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_data_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_data_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_broker_data_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace BifEnum { namespace Broker { enum DataType {
	NONE,
	BOOL,
	INT,
	COUNT,
	DOUBLE,
	STRING,
	ADDR,
	SUBNET,
	PORT,
	TIME,
	INTERVAL,
	ENUM,
	SET,
	TABLE,
	VECTOR,
}; } }
namespace zeek { namespace BifType { namespace Enum{ namespace Broker {  extern zeek::IntrusivePtr<zeek::EnumType> DataType;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Broker {  extern zeek::IntrusivePtr<zeek::RecordType> Data;  } } }}
namespace zeek { namespace BifType { namespace Record{ namespace Broker {  extern zeek::IntrusivePtr<zeek::RecordType> TableItem;  } } }}
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __data_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __data_type_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __opaque_clone_through_serialization_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_contains_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_insert_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_remove_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __set_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_contains_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_insert_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_remove_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __table_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_insert_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_replace_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_remove_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __vector_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_assign_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }
namespace zeek { namespace BifFunc { namespace Broker { extern zeek::detail::BifReturnVal __record_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }  } }

#endif
