// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/probabilistic/top-k.bif (alternative mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_top_k_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_top_k_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_probabilistic_top_k_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_init_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_add_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_get_top_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_count_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_epsilon_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_sum_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_merge_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }
namespace zeek { namespace BifFunc { extern zeek::detail::BifReturnVal topk_merge_prune_bif(zeek::detail::Frame* frame, const zeek::Args*); }  }

#endif
