// This file was automatically generated by bifcl from /pobj/zeek-6.0.5/zeek-6.0.5/src/logging/writers/sqlite/sqlite.bif (plugin mode).

#if defined(ZEEK_IN_NETVAR) || ! defined(_pobj_zeek_6_0_5_zeek_6_0_5_build_src_logging_writers_sqlite_sqlite_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _pobj_zeek_6_0_5_zeek_6_0_5_build_src_logging_writers_sqlite_sqlite_bif
#define _pobj_zeek_6_0_5_zeek_6_0_5_build_src_logging_writers_sqlite_sqlite_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek { namespace BifConst { namespace LogSQLite {  extern zeek::IntrusivePtr<zeek::StringVal> set_separator;  }  } }
namespace zeek { namespace BifConst { namespace LogSQLite {  extern zeek::IntrusivePtr<zeek::StringVal> empty_field;  }  } }
namespace zeek { namespace BifConst { namespace LogSQLite {  extern zeek::IntrusivePtr<zeek::StringVal> unset_field;  }  } }

#endif
